#!/usr/bin/env python3

"""
robohat_gateway.py

This is the OneGPIO gateway for pymata_rh

 Copyright (c) 2020 Alan Yorinks All right reserved.

 This software is free software; you can redistribute it and/or
 modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
 Version 3 as published by the Free Software Foundation; either
 or (at your option) any later version.
 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU AFFERO GENERAL PUBLIC LICENSE
 along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

"""

from python_banyan.gateway_base import GatewayBase
from pymata_rh import pymata_rh
import argparse
import signal
import sys


class RoboHatGateway(GatewayBase):
    """
    This is a good starting point for creating your own Banyan GPIO Gateway

    Search for GpioGatewayTemplate and gpio_gateway_template, and then replace
    with a names of your own making.

    Change the -l and -n options at the bottom of this file
    to be consistent with the specific board type
    """

    # noinspection PyDefaultArgument,PyRedundantParentheses
    def __init__(self, *subscriber_list, **kwargs):
        """

        :param subscriber_list: a tuple or list of topics to be subscribed to
        :param kwargs: contains the following parameters:

            back_plane_ip_address: banyan_base back_planeIP Address -
               if not specified, it will be set to the local computer
            subscriber_port: banyan_base back plane subscriber port.
               This must match that of the banyan_base backplane
            publisher_port: banyan_base back plane publisher port.
                               This must match that of the banyan_base
                               backplane
            process_name: Component identifier
            board_type: target micro-controller type ID

        """
        # initialize the parent
        super(RoboHatGateway, self).__init__(
            subscriber_list=subscriber_list,
            back_plane_ip_address=kwargs[
                'back_plane_ip_address'],
            subscriber_port=kwargs[
                'subscriber_port'],
            publisher_port=kwargs[
                'publisher_port'],
            process_name=kwargs[
                'process_name'],
            board_type=kwargs['board_type']
        )

        # instantiate pymata_rh
        self.robohat = pymata_rh.PymataRh(kwargs[
                                              'com_port'], kwargs['arduino_instance_id'])
        # start the banyan receive loop
        try:
            self.receive_loop()
        except KeyboardInterrupt:
            self.clean_up()
            sys.exit(0)

    def init_pins_dictionary(self):
        """
        The pins dictionary is an array of dictionary items that you create
        to describe each GPIO pin. In this dictionary, you can store things
        such as the pins current mode, the last value reported for an input pin
        callback method for an input pin, etc.
        """

        # not used for robohat gateway, but must be initialized.
        self.pins_dictionary = []

    def additional_banyan_messages(self, topic, payload):
        """
        This method will pass any messages not handled by this class to the
        specific gateway class. Must be overwritten by the hardware gateway
        class.
        :param topic: message topic
        :param payload: message payload
        """
        if payload['command'] == 'initialize_mpu':
            self.robohat.mpu_9250_initialize(callback=self.mpu_callback)
        elif payload['command'] == 'read_mpu':
            self.robohat.mpu_9250_read_data()
        elif payload['command'] == 'initialize_ina':
            self.robohat.ina_initialize(callback=self.ina_callback)
        elif payload['command'] == 'get_ina_bus_voltage':
            self.robohat.ina_read_bus_voltage()
        elif payload['command'] == 'get_ina_bus_current':
            self.robohat.ina_read_bus_current()
        elif payload['command'] == 'get_supply_voltage':
            self.robohat.ina_read_supply_voltage()
        elif payload['command'] == 'get_shunt_voltage':
            self.robohat.ina_read_shunt_voltage()
        elif payload['command'] == 'get_power':
            self.robohat.ina_read_power()

    def digital_write(self, topic, payload):
        """
        This method will pass any messages not handled by this class to the
        specific gateway class. Must be overwritten by the hardware gateway
        class.
        :param topic: message topic
        :param payload: message payload
        """
        self.robohat.digital_write(payload["pin"], payload['value'])

    def disable_analog_reporting(self, topic, payload):
        """
        This method will pass any messages not handled by this class to the
        specific gateway class. Must be overwritten by the hardware gateway
        class.
        :param topic: message topic
        :param payload: message payload
        """
        raise NotImplementedError

    def disable_digital_reporting(self, topic, payload):
        """
        This method will pass any messages not handled by this class to the
        specific gateway class. Must be overwritten by the hardware gateway
        class.
        :param topic: message topic
        :param payload: message payload
        """
        raise NotImplementedError

    def enable_analog_reporting(self, topic, payload):
        """
        This method will pass any messages not handled by this class to the
        specific gateway class. Must be overwritten by the hardware gateway
        class.
        :param topic: message topic
        :param payload: message payload
        """
        raise NotImplementedError

    def enable_digital_reporting(self, topic, payload):
        """
        This method will pass any messages not handled by this class to the
        specific gateway class. Must be overwritten by the hardware gateway
        class.
        :param topic: message topic
        :param payload: message payload
        """
        raise NotImplementedError

    def i2c_read(self, topic, payload):
        """
        This method will pass any messages not handled by this class to the
        specific gateway class. Must be overwritten by the hardware gateway
        class.
        :param topic: message topic
        :param payload: message payload
        """
        raise NotImplementedError

    def i2c_write(self, topic, payload):
        """
        This method will pass any messages not handled by this class to the
        specific gateway class. Must be overwritten by the hardware gateway
        class.
        :param topic: message topic
        :param payload: message payload
        """
        raise NotImplementedError

    def play_tone(self, topic, payload):
        """
        This method will pass any messages not handled by this class to the
        specific gateway class. Must be overwritten by the hardware gateway
        class.
        :param topic: message topic
        :param payload: message payload
        """
        raise NotImplementedError

    def pwm_write(self, topic, payload):
        """

        This method sets the pwm value for the selected pin.
        Call set_mode_pwm before calling this method.
        :param topic: from_robohat_gui
        :param payload: {“command”: “pwm_write”, "pin": “PIN”,
                         "tag":”TAG”,
                          “value”: “VALUE”}
        """
        self.robohat.pwm_write(payload["pin"], payload['value'])

    def servo_position(self, topic, payload):
        """
        This method will pass any messages not handled by this class to the
        specific gateway class. Must be overwritten by the hardware gateway
        class.
        :param topic: message topic
        :param payload: message payload
        """
        # self.pins_dictionary[pin][GatewayBase.PIN_MODE] = GatewayBase.SERVO_MODE
        self.robohat.servo_write(payload["pin"], payload["position"])

    def set_mode_analog_input(self, topic, payload):
        """
        This method will pass any messages not handled by this class to the
        specific gateway class. Must be overwritten by the hardware gateway
        class.
        :param topic: message topic
        :param payload: message payload
        """
        pin = payload["pin"]
        # self.pins_dictionary[pin + self.first_analog_pin][GatewayBase.PIN_MODE] = \
        #     GatewayBase.ANALOG_INPUT_MODE
        self.robohat.set_pin_mode_analog_input(pin, self.analog_input_callback)

    def set_mode_digital_input(self, topic, payload):
        """
        This method will pass any messages not handled by this class to the
        specific gateway class. Must be overwritten by the hardware gateway
        class.
        :param topic: message topic
        :param payload: message payload
        """
        pin = payload["pin"]
        # self.pins_dictionary[pin][GatewayBase.PIN_MODE] = GatewayBase.DIGITAL_INPUT_PULLUP_MODE
        self.robohat.set_pin_mode_digital_input_pullup(pin, self.digital_input_callback)

    def set_mode_digital_input_pullup(self, topic, payload):
        """
        This method will pass any messages not handled by this class to the
        specific gateway class. Must be overwritten by the hardware gateway
        class.
        :param topic: message topic
        :param payload: message payload
        """
        raise NotImplementedError

    def set_mode_digital_output(self, topic, payload):
        """
        This method will pass any messages not handled by this class to the
        specific gateway class. Must be overwritten by the hardware gateway
        class.
        :param topic: message topic
        :param payload: message payload
        """
        pin = payload["pin"]
        # self.pins_dictionary[pin][GatewayBase.PIN_MODE] = GatewayBase.DIGITAL_OUTPUT_MODE
        self.robohat.set_pin_mode_digital_output(pin)

    def set_mode_i2c(self, topic, payload):
        """
        This method will pass any messages not handled by this class to the
        specific gateway class. Must be overwritten by the hardware gateway
        class.
        :param topic: message topic
        :param payload: message payload
        """
        raise NotImplementedError

    def set_mode_pwm(self, topic, payload):
        """
        This method will pass any messages not handled by this class to the
        specific gateway class. Must be overwritten by the hardware gateway
        class.
        :param topic: message topic
        :param payload: message payload
        """
        pin = payload["pin"]
        # self.pins_dictionary[pin][GatewayBase.PIN_MODE] = GatewayBase.PWM_OUTPUT_MODE
        self.robohat.set_pin_mode_pwm_output(pin)

    def set_mode_servo(self, topic, payload):
        """
        This method will pass any messages not handled by this class to the
        specific gateway class. Must be overwritten by the hardware gateway
        class.
        :param topic: message topic
        :param payload: message payload
        """
        pin = payload["pin"]
        # self.pins_dictionary[pin][GatewayBase.PIN_MODE] = GatewayBase.SERVO_MODE
        self.robohat.set_pin_mode_servo(pin)

    def set_mode_sonar(self, topic, payload):
        """
        This method will pass any messages not handled by this class to the
        specific gateway class. Must be overwritten by the hardware gateway
        class.
        :param topic: message topic
        :param payload: message payload
        """
        raise NotImplementedError

    def set_mode_stepper(self, topic, payload):
        """
        Set the mode for the specific board.
        Must be overwritten by the hardware gateway class.
        :param topic: message topic
        :param payload: message payload
        """
        raise NotImplementedError

    def set_mode_tone(self, topic, payload):
        """
        Must be overwritten by the hardware gateway class.
        Handles a digital write
        :param topic: message topic
        :param payload: message payload
        """
        raise NotImplementedError

    def digital_read(self, pin):
        raise NotImplementedError

    def stepper_write(self, topic, payload):
        """
        Must be overwritten by the hardware gateway class.
        Handles a pwm write
        :param topic: message topic
        :param payload: message payload
        """
        raise NotImplementedError

    def analog_input_callback(self, data):
        # data = [pin mode, pin, current reported value, timestamp]
        # self.pins_dictionary[data[1] + self.arduino.first_analog_pin][GatewayBase.LAST_VALUE] = data[2]
        payload = {'report': 'analog_input', 'pin': data[1],
                   'value': data[2], 'timestamp': data[3]}
        self.publish_payload(payload, 'from_robohat_gateway')

    def digital_input_callback(self, data):
        """
        Digital input data change reported by Arduino
        :param data:
        :return:
        """
        # data = [pin mode, pin, current reported value, timestamp]
        # self.pins_dictionary[data[1]][GatewayBase.LAST_VALUE] = data[2]
        payload = {'report': 'digital_input', 'pin': data[1],
                   'value': data[2], 'timestamp': data[3]}
        self.publish_payload(payload, 'from_robohat_gateway')

    def mpu_callback(self, data):
        """
        Digital input data change reported by Arduino
        :param data:
                 index[0] = pin type - for mpu9250 the value is 16
                 index[1] = mpu address
                 index[2] = accelerometer x axis
                 index[3] = accelerometer y axis
                 index[4] = accelerometer z axis
                 index[5] = gyroscope x axis
                 index[6] = gyroscope y axis
                 index[7] = gyroscope z axis
                 index[8] = magnetometer x axis
                 index[9] = magnetometer y axis
                 index[10] = magnetometer z axis
                 index[11] = temperature
                 index[12] = timestamp
        :return:
        """
        # data = [pin mode, pin, current reported value, timestamp]
        # self.pins_dictionary[data[1]][GatewayBase.LAST_VALUE] = data[2]
        payload = {'report': 'mpu',
                   'Ax': data[2], 'Ay': data[3], 'Az': data[4],
                   'Gx': data[5], 'Gy': data[6], 'Gz': data[7],
                   'Mx': data[8], 'My': data[9], 'Mz': data[10],
                   'Temperature': data[11]
                   }
        self.publish_payload(payload, 'from_robohat_gateway')

    def ina_callback(self, data):
        payload = {}
        cb_reported_value = data[3]
        if data[2] == 0:
            payload = {'report': 'ina', 'param': 'V', 'value': cb_reported_value}
        elif data[2] == 1:
            payload = {'report': 'ina', 'param': ' A', 'value': cb_reported_value}
        elif data[2] == 2:
            payload = {'report': 'ina', 'param': 'Supply', 'value': cb_reported_value}
        elif data[2] == 3:
            payload = {'report': 'ina', 'param': 'Shunt', 'value': cb_reported_value}
        elif data[2] == 4:
            payload = {'report': 'ina', 'param': 'Power', 'value': cb_reported_value}
        self.publish_payload(payload, 'from_robohat_gateway')


def robohat_gateway():
    parser = argparse.ArgumentParser()
    parser.add_argument("-b", dest="back_plane_ip_address", default="None",
                        help="None or IP address used by Back Plane")
    parser.add_argument("-c", dest="com_port", default="None",
                        help="Use this COM port instead of auto discovery")
    parser.add_argument("-d", dest="board_type", default="None",
                        help="This parameter identifies the target GPIO "
                             "device")
    parser.add_argument("-i", dest="arduino_instance_id", default="1",
                        help="Set an Arduino Instance ID and match it in FirmataExpress")
    parser.add_argument("-g", dest="log", default="False",
                        help="Set to True to turn logging on.")
    parser.add_argument("-l", dest="subscriber_list",
                        default="to_robohat_gateway", nargs='+',
                        help="Banyan topics space delimited: topic1 topic2 "
                             "topic3")
    parser.add_argument("-n", dest="process_name", default="RoboHatGateway",
                        help="Set process name in banner")
    parser.add_argument("-p", dest="publisher_port", default='43124',
                        help="Publisher IP port")
    parser.add_argument("-s", dest="subscriber_port", default='43125',
                        help="Subscriber IP port")
    parser.add_argument("-t", dest="loop_time", default=".1",
                        help="Event Loop Timer in seconds")

    args = parser.parse_args()
    if args.back_plane_ip_address == 'None':
        args.back_plane_ip_address = None
    if args.board_type == 'None':
        args.back_plane_ip_address = None
    if args.com_port == 'None':
        args.com_port = None
    kw_options = {
        'back_plane_ip_address': args.back_plane_ip_address,
        'publisher_port': args.publisher_port,
        'subscriber_port': args.subscriber_port,
        'process_name': args.process_name,
        'loop_time': float(args.loop_time),
        'board_type': args.board_type,
        'com_port': args.com_port,
        'arduino_instance_id': int(args.arduino_instance_id)
    }
    try:
        RoboHatGateway(args.subscriber_list, **kw_options)
    except KeyboardInterrupt:
        sys.exit()


# signal handler function called when Control-C occurs
# noinspection PyShadowingNames,PyUnusedLocal,PyUnusedLocal
def signal_handler(sig, frame):
    print('Exiting Through Signal Handler')
    raise KeyboardInterrupt


# listen for SIGINT
signal.signal(signal.SIGINT, signal_handler)
signal.signal(signal.SIGTERM, signal_handler)

if __name__ == '__main__':
    # replace with name of function you defined above
    robohat_gateway()



"""
 Copyright (c) 2019 Alan Yorinks All rights reserved.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
 Version 3 as published by the Free Software Foundation; either
 or (at your option) any later version.
 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU AFFERO GENERAL PUBLIC LICENSE
 along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""
import atexit
import signal
import subprocess
import sys
import time

import psutil


# import webbrowser


class S3RP:
    """
    This class starts the Banyan server to support the Scratch 3 OneGPIO Raspberry Pi
    extension for local mode. That means the browser and components are all executing
    on a single RPi.

    It will start the backplane, Raspberry Pi gateway and websocket gateway.
    """

    def __init__(self):
        """
        Prepare for launching the rpi extension
        """

        self.proc_bp = None
        self.proc_awg = None
        self.proc_hwg = None

        self.skip_backplane = False

        # start backplane
        self.proc_bp = self.start_backplane()
        if self.proc_bp:
            print('backplane started')

        else:
            print('backplane start failed - exiting')
            sys.exit(0)

        self.proc_awg = self.start_wsgw()
        if self.proc_awg:
            print('Websocket Gateway started')
        else:
            print('WebSocket Gateway start failed - exiting')
            sys.exit(0)

        # start rpi gateway
        self.proc_hwg = self.start_rp_gw()
        if self.proc_hwg:
            print('RPi Pico Gateway started ')
            print('To exit this program, press Control-c')

        else:
            print('RPi Gateway start failed - exiting')
            sys.exit(0)

        atexit.register(self.killall)
        # webbrowser.open('https://mryslab.github.io/s3onegpio/', new=1)

        while True:
            try:
                if not self.skip_backplane:
                    if self.proc_bp.poll() is not None:
                        self.proc_bp = None
                        print('backplane exited...')
                        self.killall()
                if self.proc_awg.poll() is not None:
                    self.proc_awg = None
                    print('Websocket Gateway exited...')
                    self.killall()
                if self.proc_hwg.poll() is not None:
                    self.proc_hwg = None
                    print('RPi Pico Gateway exited.')
                    self.killall()

                # allow some time between polls
                time.sleep(.4)
            except KeyboardInterrupt:
                self.killall()

    def killall(self):
        """
        Kill all running processes
        """

        # check for missing processes
        if self.proc_bp:
            try:
                if sys.platform.startswith('win32'):
                    subprocess.run(['taskkill', '/F', '/t', '/PID', str(self.proc_bp.pid)],
                                   creationflags=subprocess.CREATE_NEW_PROCESS_GROUP |
                                                 subprocess.CREATE_NO_WINDOW
                                   )
                else:
                    self.proc_bp.kill()
                self.proc_bp = None
            except:
                pass
        if self.proc_awg:
            try:
                if sys.platform.startswith('win32'):
                    subprocess.run(['taskkill', '/F', '/t', '/pid', str(self.proc_awg.pid)],
                                   creationflags=subprocess.CREATE_NEW_PROCESS_GROUP |
                                                 subprocess.CREATE_NO_WINDOW
                                   )
                else:
                    self.proc_awg.kill()
                self.proc_awg = None
            except:
                pass
        if self.proc_hwg:
            try:
                if sys.platform.startswith('win32'):
                    subprocess.run(['taskkill', '/F', '/t', '/PID', str(self.proc_hwg.pid)],
                                   creationflags=subprocess.CREATE_NEW_PROCESS_GROUP |
                                                 subprocess.CREATE_NO_WINDOW
                                   )
                else:
                    self.proc_hwg.kill()
                self.proc_hwg = None
            except:
                pass
        sys.exit(0)

    def start_backplane(self):
        """
        Start the backplane
        """

        # check to see if the backplane is already running
        try:
            for proc in psutil.process_iter(attrs=['pid', 'name']):
                if 'backplane' in proc.info['name']:
                    self.skip_backplane = True
                    # its running - return its pid
                    return proc
        except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
            pass

        # backplane is not running, so start one
        if sys.platform.startswith('win32'):
            return subprocess.Popen(['backplane'],
                                    creationflags=subprocess.CREATE_NEW_PROCESS_GROUP |
                                                  subprocess.CREATE_NO_WINDOW)
        else:
            return subprocess.Popen(['backplane'],
                                    stdin=subprocess.PIPE, stderr=subprocess.PIPE,
                                    stdout=subprocess.PIPE)

    def start_wsgw(self):
        """
        Start the websocket gateway
        """
        if sys.platform.startswith('win32'):
            return subprocess.Popen(['wsgw', '-i', '9006'],
                                    creationflags=subprocess.CREATE_NEW_PROCESS_GROUP
                                                  |
                                                  subprocess.CREATE_NO_WINDOW)
        else:
            return subprocess.Popen(['wsgw', '-i', '9006'],
                                    stdin=subprocess.PIPE, stderr=subprocess.PIPE,
                                    stdout=subprocess.PIPE)

    def start_rp_gw(self):
        """
        Start the rpi pico gateway
        """
        if sys.platform.startswith('win32'):
            return subprocess.Popen(['rpgw'], creationflags=subprocess.CREATE_NEW_PROCESS_GROUP |
                                                             subprocess.CREATE_NO_WINDOW)
        else:
            return subprocess.Popen(['rpgw'], stdin=subprocess.PIPE, stderr=subprocess.PIPE,
                                    stdout=subprocess.PIPE)


def signal_handler(sig, frame):
    print('Exiting Through Signal Handler')
    raise KeyboardInterrupt


def s3rpx():
    """
    Start the extension
    :return:
    """
    S3RP()


# listen for SIGINT
signal.signal(signal.SIGINT, signal_handler)
signal.signal(signal.SIGTERM, signal_handler)

if __name__ == '__main__':
    # replace with name of function you defined above
    s3rpx()



"""
 This is the Python Banyan Hardware Gateway for the
 Adafruit Circuit Playground Express.

 Copyright (c) 2020 Alan Yorinks All right reserved.

 Python Banyan is free software; you can redistribute it and/or
 modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
 Version 3 as published by the Free Software Foundation; either
 or (at your option) any later version.
 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU AFFERO GENERAL PUBLIC LICENSE
 along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

"""
import argparse
import atexit
import logging
import math
import os
import pathlib
import signal
import sys
import threading
import time

import serial
from pymata_cpx.pymata_cpx import PyMataCpx
from python_banyan.gateway_base import GatewayBase


# noinspection PyMethodMayBeStatic
class CpxGateway(GatewayBase, threading.Thread):
    """
    This class is the interface class for the Circuit Playground
    Express supporting Scratch 3.
    """

    def __init__(self, *subscriber_list, back_plane_ip_address=None, subscriber_port='43125',
                 publisher_port='43124', process_name='CpxGateway',
                 publisher_topic=None, log=False):
        """
        :param subscriber_list: a tuple or list of subscription topics.
        :param back_plane_ip_address:
        :param subscriber_port:
        :param publisher_port:
        :param process_name:
        """

        # initialize parent
        super(CpxGateway, self).__init__(subscriber_list=subscriber_list,
                                         back_plane_ip_address=back_plane_ip_address,
                                         subscriber_port=subscriber_port,
                                         publisher_port=publisher_port, process_name=process_name)
        self.log = log
        if self.log:
            fn = str(pathlib.Path.home()) + "/cpxgw.log"
            self.logger = logging.getLogger(__name__)
            logging.basicConfig(filename=fn, filemode='w', level=logging.DEBUG)
            sys.excepthook = self.my_handler

        self.publisher_topic = publisher_topic
        self.cpx = PyMataCpx()
        atexit.register(self.shutdown)

        # hold the time of the last analog data to be received.
        # use to determine if connectivity is gone.
        self.last_analog_data_time = None

        # start up all the sensors
        self.cpx.cpx_accel_start(self.tilt_callback)
        self.cpx.cpx_button_a_start(self.switch_callback)
        self.cpx.cpx_button_b_start(self.switch_callback)
        self.cpx.cpx_slide_switch_start(self.switch_callback)

        self.cpx.cpx_light_sensor_start(self.analog_callback)
        self.cpx.cpx_microphone_start(self.analog_callback)
        self.cpx.cpx_temperature_start(self.analog_callback)
        for touch_pad in range(1, 8):
            self.cpx.cpx_cap_touch_start(touch_pad, self.touchpad_callback)

        threading.Thread.__init__(self)
        self.daemon = True

        # start the watchdog thread
        self.start()
        # start the banyan receive loop
        try:
            self.receive_loop()
        except:
            pass
        # except KeyboardInterrupt:
        # except KeyboardInterrupt:
            # self.cpx.cpx_close_and_exit()
            # sys.exit(0)
        #     os._exit(1)


    def init_pins_dictionary(self):
        pass

    def play_tone(self, topic, payload):
        """
        This method plays a tone on a piezo device connected to the selected
        pin at the frequency and duration requested.
        Frequency is in hz and duration in milliseconds.

        Call set_mode_tone before using this method.
        :param topic: message topic
        :param payload: {"command": "play_tone", "pin": “PIN”, "tag": "TAG",
                         “freq”: ”FREQUENCY”, duration: “DURATION”}
        """
        self.cpx.cpx_tone(payload['freq'], payload['duration'])

    def additional_banyan_messages(self, topic, payload):
        if payload['command'] == 'pixel':
            self.set_pixel(payload)

    def set_pixel(self, payload):
        self.cpx.cpx_pixel_set(payload['pixel'], payload['red'],
                               payload['green'], payload['blue'])
        self.cpx.cpx_pixels_show()

    def digital_write(self, topic, payload):
        """
        This method performs a digital write to the board LED
        :param topic: message topic
        :param payload: {"command": "digital_write", "pin": “PIN”, "value": “VALUE”}
        """
        if payload['value']:
            self.cpx.cpx_board_light_on()
        else:
            self.cpx.cpx_board_light_off()

    # The CPX sensor callbacks

    def tilt_callback(self, data):
        """
        Report the tilt of the express board

        Take the raw xyz data and transform it to
        positional strings.
        :param data: data [0] = data mode 32 is analog.
                     data[1] = the pin number - this is a pseudo pin number
                     data[2] = x value
                     data[3] = y value
                     data[4] = z value
        """
        x = data[2]
        y = data[3]
        z = data[4]

        # Convert raw Accelerometer values to degrees
        x_angle = int((math.atan2(y, z) + math.pi) * (180 / math.pi))
        y_angle = int((math.atan2(z, x) + math.pi) * (180 / math.pi))

        h = v = -1

        if 175 < x_angle < 185 and 265 < y_angle < 275:
            h = v = 0  # 'flat'

        elif h or v:
            if 180 <= x_angle <= 270:
                v = 1  # up

            elif 90 <= x_angle <= 180:
                v = 2  # down

            if 180 <= y_angle <= 270:
                h = 3  # left

            elif 270 <= y_angle <= 360:
                h = 4  # right

        payload = {'report': 'tilted', 'value': [v, h]}
        self.publish_payload(payload, 'from_cpx_gateway')

    def switch_callback(self, data):
        """
        This handles switches a, b, and slide
        :param data: data[1] - a=4 b=5, slice=7,
        """
        if data[1] == 4:
            switch = 'a'
        elif data[1] == 5:
            switch = 'b'
        else:
            # 0 = right, 1 = left
            switch ='slide'

        payload = {'report': switch, 'value': data[2]}
        self.publish_payload(payload, 'from_cpx_gateway')

    def analog_callback(self, data):
        """
        This handles the light, temperature and sound sensors.

        It also sets up a "watchdog timer" and if there is no activity
        for > 1 second will exit.


        :param data: data[1] - 8 = light, temp = 9, 10 = sound,
        """

        self.last_analog_data_time = time.time()

        if data[1] == 8:
            sensor = 'light'
        elif data[1] == 9:
            sensor = 'temp'
        else:
            sensor = 'sound'
        payload = {'report': sensor, 'value': round(data[2], 2)}
        self.publish_payload(payload, 'from_cpx_gateway')

    def touchpad_callback(self, data):
        """
        Build and send a banyan message for the pad and value
        :param data: data[1] = touchpad and data[2] = boolean value
        """
        payload = {'report': 'touch' + str(data[1]), 'value': int(data[2])}
        self.publish_payload(payload, 'from_cpx_gateway')

    def shutdown(self):
        try:
            self.cpx.cpx_close_and_exit()
            sys.exit(0)
        except serial.serialutil.SerialException:
            pass

    def my_handler(self, xtype, value, tb):
        """
        for logging uncaught exceptions
        :param xtype:
        :param value:
        :param tb:
        :return:
        """
        self.logger.exception("Uncaught exception: {0}".format(str(value)))

    def run(self):
        if not self.last_analog_data_time:
            self.last_analog_data_time = time.time()
        while True:
            if time.time() - self.last_analog_data_time > 1.0:
                print('Watchdog timed out - exiting.')
                os._exit(1)

            time.sleep(1)


def cpx_gateway():
    parser = argparse.ArgumentParser()
    parser.add_argument("-b", dest="back_plane_ip_address", default="None",
                        help="None or IP address used by Back Plane")
    parser.add_argument("-c", dest="com_port", default="None",
                        help="Use this COM port instead of auto discovery")
    parser.add_argument("-l", dest="log", default="False",
                        help="Set to True to turn logging on.")
    parser.add_argument("-m", dest="subscriber_list",
                        default="to_cpx_gateway", nargs='+',
                        help="Banyan topics space delimited: topic1 topic2 topic3")
    parser.add_argument("-n", dest="process_name",
                        default="CpxGateway", help="Set process name in "
                                                   "banner")
    parser.add_argument("-p", dest="publisher_port", default='43124',
                        help="Publisher IP port")
    parser.add_argument("-r", dest="publisher_topic",
                        default="from_cpx_gateway", help="Report topic")
    parser.add_argument("-s", dest="subscriber_port", default='43125',
                        help="Subscriber IP port")

    args = parser.parse_args()

    subscriber_list = args.subscriber_list

    kw_options = {
        'publisher_port': args.publisher_port,
        'subscriber_port': args.subscriber_port,
        'process_name': args.process_name,
        'publisher_topic': args.publisher_topic
    }

    if args.back_plane_ip_address != 'None':
        kw_options['back_plane_ip_address'] = args.back_plane_ip_address

    if args.com_port != 'None':
        kw_options['com_port'] = args.com_port

    log = args.log.lower()
    if log == 'false':
        log = False
    else:
        log = True

    kw_options['log'] = log

    cpx = CpxGateway(subscriber_list, **kw_options)


# signal handler function called when Control-C occurs
# noinspection PyShadowingNames,PyUnusedLocal
def signal_handler(sig, frame):
    print('Exiting Through Signal Handler')
    raise



# listen for SIGINT
signal.signal(signal.SIGINT, signal_handler)
signal.signal(signal.SIGTERM, signal_handler)


if __name__ == '__main__':
    cpx_gateway()



"""
 Copyright (c) 2018-2021 Alan Yorinks All rights reserved.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
 Version 3 as published by the Free Software Foundation; either
 or (at your option) any later version.
 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU AFFERO GENERAL PUBLIC LICENSE
 along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

import argparse
import asyncio
import signal
import sys
import time

from python_banyan.gateway_base_aio import GatewayBaseAIO
from telemetrix_aio.telemetrix_aio import TelemetrixAIO


# noinspection PyAbstractClass,PyMethodMayBeStatic,PyRedundantParentheses,DuplicatedCode,GrazieInspection
class Esp8266Gateway(GatewayBaseAIO):
    # This class implements the GatewayBase interface adapted for asyncio.
    # It supports ESP8266 boards, tested with nodemcu.

    # NOTE: This class requires the use of Python 3.8 or above

    # serial_port = None

    def __init__(self, *subscriber_list, back_plane_ip_address=None,
                 subscriber_port='43125',
                 publisher_port='43124', process_name='Esp8266Gateway',
                 event_loop=None):
        """
        Set up the gateway for operation

        :param subscriber_list: a tuple or list of subscription topics.
        :param back_plane_ip_address: ip address of backplane or none if local
        :param subscriber_port: backplane subscriber port
        :param publisher_port: backplane publisher port
        :param process_name: name to display on the console
        :param event_loop: optional parameter to pass in an asyncio
                           event loop

        """

        self.pin_info = {}
        self.gpio_pins = [4, 5, 12, 13, 14, 15]

        self.back_plane_ip_address = back_plane_ip_address
        self.publisher_port = publisher_port
        self.subscriber_port = subscriber_port

        self.subscriber_list = subscriber_list

        # set the event loop to be used. accept user's if provided
        self.event_loop = event_loop

        self.connection_socket = False

        # instantiate TelemetrixAIO
        # if user want to pass in a com port, then pass it in
        self.esp = TelemetrixAIO(autostart=False, loop=self.event_loop, com_port=None)

        # Initialize the parent
        super(Esp8266Gateway, self).__init__(subscriber_list=subscriber_list,
                                             event_loop=self.event_loop,
                                             back_plane_ip_address=back_plane_ip_address,
                                             subscriber_port=subscriber_port,
                                             publisher_port=publisher_port,
                                             process_name=process_name,
                                             )

        time.sleep(.5)

    def init_pins_dictionary(self):
        """
        This method will initialize the pins dictionary contained
        in gateway base parent class. This method is called by
        the gateway base parent in its init method.

        NOTE: that this a a non-asyncio method.
        """
        # pins data structure
        # pins supported: GPIO4, GPIO5, GPIO12, GPIO13, GPIO14, GPIO15
        # modes - IN - PULL_UP, OUT - OPEN_DRAIN, PWM
        # irqs - Pin.IRQ_RISING | Pin.IRQ_FALLING

        # build a status table for the pins
        for x in self.gpio_pins:
            entry = {'mode': None, 'pull_up': False, 'drain': False,
                     'irq': None, 'duty': None, 'freq': None, 'count': 0,
                     'value': 0}
            self.pin_info[x] = entry

    async def main(self):
        # call the inherited begin method located in banyan_base_aio
        await self.begin()

        # sit in an endless loop to receive protocol messages
        while True:
            await self.receive_loop()

    async def additional_banyan_messages(self, topic, payload):
        if payload['command'] == 'ip_address':
            # start up telemetrix-aio
            if not self.connection_socket:
                self.connection_socket = True
                self.esp.ip_port = 31335
                self.esp.ip_address = payload['address']
                await self.esp.start_aio()
                await asyncio.sleep(1)

    # The following methods and are called
    # by the gateway base class in its incoming_message_processing
    # method. They overwrite the default methods in the gateway_base.

    async def digital_write(self, topic, payload):
        """
        This method performs a digital write
        :param topic: message topic
        :param payload: {"command": "digital_write", "pin": “PIN”, "value": “VALUE”}
        """

        await self.esp.digital_write(payload["pin"], payload['value'])

    async def disable_analog_reporting(self, topic, payload):
        """
        This method disables analog input reporting for the selected pin.
        :param topic: message topic
        :param payload: {"command": "disable_analog_reporting", "pin": “PIN”, "tag": "TAG"}
        """
        await self.esp.disable_analog_reporting(payload["pin"])

    async def disable_digital_reporting(self, topic, payload):
        """
        This method disables digital input reporting for the selected pin.

        :param topic: message topic
        :param payload: {"command": "disable_digital_reporting", "pin": “PIN”, "tag": "TAG"}
        """
        await self.esp.disable_digital_reporting(payload["pin"])

    async def enable_analog_reporting(self, topic, payload):
        """
        This method enables analog input reporting for the selected pin.
        :param topic: message topic
        :param payload:  {"command": "enable_analog_reporting", "pin": “PIN”, "tag": "TAG"}
        """
        await self.esp.enable_analog_reporting(payload["pin"])

    async def enable_digital_reporting(self, topic, payload):
        """
        This method enables digital input reporting for the selected pin.
        :param topic: message topic
        :param payload: {"command": "enable_digital_reporting", "pin": “PIN”, "tag": "TAG"}
        """
        await self.esp.enable_digital_reporting(payload["pin"])

    async def i2c_read(self, topic, payload):
        """
        This method will perform an i2c read by specifying the i2c
        device address, i2c device register and the number of bytes
        to read.

        Call set_mode_i2c first to establish the pins for i2c operation.

        :param topic: message topic
        :param payload: {"command": "i2c_read", "pin": “PIN”, "tag": "TAG",
                         "addr": “I2C ADDRESS, "register": “I2C REGISTER”,
                         "number_of_bytes": “NUMBER OF BYTES”}
        :return via the i2c_callback method
        """

        await self.esp.i2c_read(payload['addr'],
                                payload['register'],
                                payload['number_of_bytes'], callback=self.i2c_callback)

    async def i2c_write(self, topic, payload):
        """
        This method will perform an i2c write for the i2c device with
        the specified i2c device address, i2c register and a list of byte
        to write.

        Call set_mode_i2c first to establish the pins for i2c operation.

        :param topic: message topic
        :param payload: {"command": "i2c_write", "pin": “PIN”, "tag": "TAG",
                         "addr": “I2C ADDRESS, "register": “I2C REGISTER”,
                         "data": [“DATA IN LIST FORM”]}
        """
        await self.esp.i2c_write(payload['addr'], payload['data'])

    async def pwm_write(self, topic, payload):
        """
        This method sets the pwm value for the selected pin.
        Call set_mode_pwm before calling this method.
        :param topic: message topic
        :param payload: {“command”: “pwm_write”, "pin": “PIN”,
                         "tag":”TAG”,
                          “value”: “VALUE”}
        """
        await self.esp.analog_write(payload["pin"], payload['value'])

    async def servo_position(self, topic, payload):
        """
        This method will set a servo's position in degrees.
        Call set_mode_servo first to activate the pin for
        servo operation.

        :param topic: message topic
        :param payload: {'command': 'servo_position',
                         "pin": “PIN”,'tag': 'servo',
                        “position”: “POSITION”}
        """
        await self.esp.servo_write(payload["pin"], payload["position"])

    async def set_mode_analog_input(self, topic, payload):
        """
        This method sets a GPIO pin as analog input.
        :param topic: message topic
        :param payload: {"command": "set_mode_analog_input", "pin": “PIN”, "tag":”TAG” }
        """
        pin = payload["pin"]
        await self.esp.set_pin_mode_analog_input(pin, callback=self.analog_input_callback)

    async def set_mode_digital_input(self, topic, payload):
        """
        This method sets a pin as digital input.
        :param topic: message topic
        :param payload: {"command": "set_mode_digital_input", "pin": “PIN”, "tag":”TAG” }
        """
        pin = payload["pin"]

        await self.esp.set_pin_mode_digital_input(pin, self.digital_input_callback)
        await asyncio.sleep(0.4)

    async def set_mode_digital_input_pullup(self, topic, payload):
        """
        This method sets a pin as digital input with pull up enabled.
        :param topic: message topic
        :param payload: message payload
        """
        pin = payload["pin"]

        await self.esp.set_pin_mode_digital_input_pullup(pin, self.digital_input_callback)

    async def set_mode_digital_output(self, topic, payload):
        """
        This method sets a pin as a digital output pin.
        :param topic: message topic
        :param payload: {"command": "set_mode_digital_output", "pin": PIN, "tag":”TAG” }
        """
        pin = payload["pin"]

        await self.esp.set_pin_mode_digital_output(pin)

    async def set_mode_i2c(self, topic, payload):
        """
        This method sets up the i2c pins for i2c operations.
        :param topic: message topic
        :param payload: {"command": "set_mode_i2c"}
        """
        # self.pins_dictionary[200][GatewayBaseAIO.PIN_MODE] = GatewayBaseAIO.I2C_MODE
        await self.esp.set_pin_mode_i2c()

    async def set_mode_pwm(self, topic, payload):
        """
        This method sets a GPIO pin capable of PWM for PWM operation.
        :param topic: message topic
        :param payload: {"command": "set_mode_pwm", "pin": “PIN”, "tag":”TAG” }
        """
        pin = payload["pin"]

        await self.esp.set_pin_mode_analog_output(pin)

    async def set_mode_servo(self, topic, payload):
        """
        This method establishes a GPIO pin for servo operation.
        :param topic: message topic
        :param payload: {"command": "set_mode_servo", "pin": “PIN”, "tag":”TAG” }
        """
        pin = payload["pin"]

        await self.esp.set_pin_mode_servo(pin)

    async def set_mode_sonar(self, topic, payload):
        """
        This method sets the trigger and echo pins for sonar operation.
        :param topic: message topic
        :param payload: {"command": "set_mode_sonar", "trigger_pin": “PIN”, "tag":”TAG”
                         "echo_pin": “PIN”"tag":”TAG” }
        """

        trigger = payload["trigger_pin"]
        echo = payload["echo_pin"]

        await self.esp.set_pin_mode_sonar(trigger, echo, self.sonar_callback)

    # Callbacks
    async def digital_input_callback(self, data):
        """
        Digital input data change reported by ESP
        :param data:
        :return:
        """
        payload = {'report': 'digital_input', 'pin': data[1],
                   'value': data[2], 'timestamp': data[3]}
        await self.publish_payload(payload, 'from_esp8266_gateway')

    async def analog_input_callback(self, data):
        payload = {'report': 'analog_input', 'pin': data[1],
                   'value': data[2], 'timestamp': data[3]}
        await self.publish_payload(payload, 'from_esp8266_gateway')

    async def i2c_callback(self, data):
        """
        Analog input data change reported by ESP

        :param data:
        :return:
        """
        # creat a string representation of the data returned
        report = ', '.join([str(elem) for elem in data])
        payload = {'report': 'i2c_data', 'value': report}
        await self.publish_payload(payload, 'from_esp8266_gateway')

    async def sonar_callback(self, data):
        """
        Sonar data change reported by ESP

        :param data:
        :return:
        """
        payload = {'report': 'sonar_data', 'value': data[2]}
        await self.publish_payload(payload, 'from_esp8266_gateway')


# noinspection DuplicatedCode
def esp8266_gateway():
    # allow user to bypass the IP address auto-discovery. This is necessary if the component resides on a computer
    # other than the computing running the backplane.

    parser = argparse.ArgumentParser()
    parser.add_argument("-b", dest="back_plane_ip_address", default="None",
                        help="None or IP address used by Back Plane")
    parser.add_argument("-c", dest="com_port", default="None",
                        help="Use this COM port instead of auto discovery")
    parser.add_argument("-m", dest="subscriber_list",
                        default="to_esp8266_gateway", nargs='+',
                        help="Banyan topics space delimited: topic1 topic2 topic3")
    parser.add_argument("-n", dest="process_name",
                        default="ESP8266Gateway", help="Set process name in "
                                                       "banner")
    parser.add_argument("-p", dest="publisher_port", default='43124',
                        help="Publisher IP port")
    parser.add_argument("-r", dest="publisher_topic",
                        default="from_rpi_gpio", help="Report topic")
    parser.add_argument("-s", dest="subscriber_port", default='43125',
                        help="Subscriber IP port")

    args = parser.parse_args()

    subscriber_list = args.subscriber_list

    kw_options = {
        'publisher_port': args.publisher_port,
        'subscriber_port': args.subscriber_port,
        'process_name': args.process_name,
    }
    if args.back_plane_ip_address != 'None':
        kw_options['back_plane_ip_address'] = args.back_plane_ip_address

    if args.com_port != 'None':
        kw_options['com_port'] = args.com_port

    # get the event loop
    # this is for python 3.8
    if sys.platform == 'win32':
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

    loop = asyncio.get_event_loop()

    # replace with the name of your class
    app = Esp8266Gateway(subscriber_list, **kw_options, event_loop=loop)
    try:
        loop.run_until_complete(app.main())
    except (KeyboardInterrupt, asyncio.CancelledError, RuntimeError):
        loop.stop()
        loop.close()
        sys.exit(0)


# signal handler function called when Control-C occurs
# noinspection PyShadowingNames,PyUnusedLocal
def signal_handler(sig, frame):
    print('Exiting Through Signal Handler')
    raise KeyboardInterrupt


# listen for SIGINT
signal.signal(signal.SIGINT, signal_handler)
signal.signal(signal.SIGTERM, signal_handler)

if __name__ == '__main__':
    esp8266_gateway()



#!/usr/bin/env python

import pigpio
import time

# based on an example provided with the pigpio library

class Sonar:
    """
    This class encapsulates a sonar device with separate
    trigger and echo pins.

    A pulse on the trigger initiates the sonar ping and shortly
    afterwards a sonar pulse is transmitted and the echo pin
    goes high.  The echo pins stays high until a sonar echo is
    received (or the response times-out).  The time between
    the high and low edges indicates the sonar round trip time.
    """

    def __init__(self, pi, trigger, echo):
        """
        The class is instantiated with the Pi to use and the
        gpios connected to the trigger and echo pins.
        """
        self.pi = pi
        self._trig = trigger
        self._echo = echo

        self._ping = False
        self._high = None
        self._time = None

        self._triggered = False

        self._trig_mode = pi.get_mode(self._trig)
        self._echo_mode = pi.get_mode(self._echo)

        pi.set_mode(self._trig, pigpio.OUTPUT)
        pi.set_mode(self._echo, pigpio.INPUT)

        self._cb = pi.callback(self._trig, pigpio.EITHER_EDGE, self._cbf)
        self._cb = pi.callback(self._echo, pigpio.EITHER_EDGE, self._cbf)

        self._inited = True

    def _cbf(self, gpio, level, tick):
        if gpio == self._trig:
            if level == 0:  # trigger sent
                self._triggered = True
                self._high = None
        else:
            if self._triggered:
                if level == 1:
                    self._high = tick
                else:
                    if self._high is not None:
                        self._time = tick - self._high
                        self._high = None
                        self._ping = True

    def read(self):
        """
        Triggers a reading.  The returned reading is the number
        of microseconds for the sonar round-trip.

        round trip cms = round trip time / 1000000.0 * 34030
        """
        if self._inited:
            self._ping = False
            self.pi.gpio_trigger(self._trig)
            start = time.time()
            while not self._ping:
                if (time.time() - start) > 5.0:
                    return 20000
                time.sleep(0.001)
            return self._time
        else:
            return None

    def cancel(self):
        """
        Cancels the ranger and returns the gpios to their
        original mode.
        """
        if self._inited:
            self._inited = False
            self._cb.cancel()
            self.pi.set_mode(self._trig, self._trig_mode)
            self.pi.set_mode(self._echo, self._echo_mode)


# if __name__ == "__main__":
#
#     import time
#
#     import pigpio
#
#     import sonar_trigger_echo
#
#     pi = pigpio.pi()
#
#     sonar = sonar_trigger_echo.Sonar(pi, 23, 18)
#
#     end = time.time() + 600.0
#
#     r = 1
#     while time.time() < end:
#         stime = sonar.read()
#         distance = stime/ 29 / 2
#         distance = round(distance, 2)
#         print("{} {}".format(r, distance))
#         r += 1
#         time.sleep(0.03)
#
#     sonar.cancel()
#
#     pi.stop()



#!/usr/bin/env python3

"""
 This is the Python Banyan GUI that communicates with
 the Raspberry Pi Banyan Gateway

 Copyright (c) 2019 Alan Yorinks All right reserved.

 Python Banyan is free software; you can redistribute it and/or
 modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
 Version 3 as published by the Free Software Foundation; either
 or (at your option) any later version.
 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU AFFERO GENERAL PUBLIC LICENSE
 along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

"""

import argparse
import asyncio
import datetime
import json
import logging
import pathlib
import signal
import sys
import websockets

from python_banyan.banyan_base_aio import BanyanBaseAIO


class WsGateway(BanyanBaseAIO):
    """
    This class is a gateway between a websocket client and the
    Banyan network.

    NOTE: This class requires Python 3.7 or above.

    It implements a websocket server. A websocket client, upon
    connection, must send an id message e.g.: {"id": "to_arduino"}.

    The id will be used as the topic to publish data to the banyan
    network.
    """

    def __init__(self, subscription_list, back_plane_ip_address=None,
                 subscriber_port='43125',
                 publisher_port='43124', process_name='WebSocketGateway',
                 event_loop=None, server_ip_port=9000, log=False):
        """
        These are all the normal base class parameters
        :param subscription_list:
        :param back_plane_ip_address:
        :param subscriber_port:
        :param publisher_port:
        :param process_name:
        :param event_loop:
        """

        # set up logging if requested
        self.log = log
        self.event_loop = event_loop

        # a kludge to shutdown the socket on control C
        self.wsocket = None

        if self.log:
            fn = str(pathlib.Path.home()) + "/wsgw.log"
            self.logger = logging.getLogger(__name__)
            logging.basicConfig(filename=fn, filemode='w', level=logging.DEBUG)
            sys.excepthook = self.my_handler

        # initialize the base class
        super(WsGateway, self).__init__(subscriber_list=subscription_list,
                                        back_plane_ip_address=back_plane_ip_address,
                                        subscriber_port=subscriber_port,
                                        publisher_port=publisher_port,
                                        process_name=process_name,
                                        event_loop=self.event_loop)

        # save the server port number
        self.server_ip_port = server_ip_port

        # array of active sockets
        self.active_sockets = []
        try:
            self.start_server = websockets.serve(self.wsg,
                                                 '0.0.0.0',
                                                 self.server_ip_port)
            print('WebSocket using: ' + self.back_plane_ip_address
                  + ':' + self.server_ip_port)
            # start the websocket server and call the main task, wsg
            self.event_loop.run_until_complete(self.start_server)
            self.event_loop.create_task(self.wakeup())
            self.event_loop.run_forever()
        except (websockets.exceptions.ConnectionClosed,
                RuntimeError,
                KeyboardInterrupt):
            if self.log:
                logging.exception("Exception occurred", exc_info=True)
            self.event_loop.stop()
            self.event_loop.close()

    async def wakeup(self):
        while True:
            try:
                await asyncio.sleep(1)
            except KeyboardInterrupt:
                for task in asyncio.Task.all_tasks():
                    task.cancel()
                await self.wsocket.close()
                self.event_loop.stop()
                self.event_loop.close()
                sys.exit(0)


    async def wsg(self, websocket, path):
        """
        This method handles connections and will be used to send
        messages to the client
        :param websocket: websocket for connected client
        :param path: required, but unused
        :return:
        """
        self.wsocket = websocket
        # start up banyan
        await self.begin()

        # wait for a connection
        try:
            data = await websocket.recv()
        except websockets.exceptions.ConnectionClosedOK:
            pass

        # expecting an id string from client
        data = json.loads(data)

        # if id field not present then raise an exception
        try:
            id_string = data['id']
        except KeyError:
            print('Client did not provide an ID string')
            raise

        # create a subscriber string from the id
        subscriber_string = id_string.replace('to', 'from')

        # subscribe to that topic
        await self.set_subscriber_topic(subscriber_string)

        # add an entry into the active_sockets table
        entry = {websocket: 'to_banyan_topic', subscriber_string: websocket}
        self.active_sockets.append(entry)

        # create a task to receive messages from the client
        await asyncio.create_task(self.receive_data(websocket, data['id']))

    async def receive_data(self, websocket, publisher_topic):
        """
        This method processes a received WebSocket command message
        and translates it to a Banyan command message.
        :param websocket: The currently active websocket
        :param publisher_topic: The publishing topic
        """
        while True:
            try:
                data = await websocket.recv()
                data = json.loads(data)
            except (websockets.exceptions.ConnectionClosed, TypeError):
                # remove the entry from active_sockets
                # using a list comprehension
                self.active_sockets = [entry for entry in self.active_sockets if websocket not in entry]
                break

            await self.publish_payload(data, publisher_topic)

    async def incoming_message_processing(self, topic, payload):
        """
        This method converts the incoming messages to ws messages
        and sends them to the ws client

        :param topic: Message Topic string.

        :param payload: Message Data.
        """
        if payload['report'] == 'panic':
            # close the sockets if in panic mode
            for socket in self.active_sockets:
                if topic in socket.keys():
                    pub_socket = socket[topic]
                    await pub_socket.close()

        if 'timestamp' in payload:
            timestamp = datetime.datetime.fromtimestamp(payload['timestamp']).strftime('%Y-%m-%d %H:%M:%S')
            payload['timestamp'] = timestamp

        ws_data = json.dumps(payload)

        # find the websocket of interest by looking for the topic in
        # active_sockets
        for socket in self.active_sockets:
            if topic in socket.keys():
                pub_socket = socket[topic]
                await pub_socket.send(ws_data)

    def my_handler(self, the_type, value, tb):
        """
        For logging uncaught exceptions
        :param the_type:
        :param value:
        :param tb:
        :return:
        """
        self.logger.exception("Uncaught exception: {0}".format(str(value)))


def ws_gateway():
    # allow user to bypass the IP address auto-discovery. This is necessary if the component resides on a computer
    # other than the computing running the backplane.

    parser = argparse.ArgumentParser()
    parser.add_argument("-b", dest="back_plane_ip_address", default="None",
                        help="None or IP address used by Back Plane")
    # allow the user to specify a name for the component and have it shown on the console banner.
    # modify the default process name to one you wish to see on the banner.
    # change the default in the derived class to set the name
    parser.add_argument("-m", dest="subscription_list", default="from_arduino_gateway, "
                                                                "from_esp8266_gateway, "
                                                                "from_rpi_gateway, "
                                                                "from_microbit_gateway"
                                                                "from_picoboard_gateway"
                                                                "from_cpx_gateway",
                                                                nargs='+',
                        help="A space delimited list of topics")
    parser.add_argument("-i", dest="server_ip_port", default="9000",
                        help="Set the WebSocket Server IP Port number")
    parser.add_argument("-l", dest="log", default="False",
                        help="Set to True to turn logging on.")
    parser.add_argument("-n", dest="process_name", default="WebSocket Gateway",
                        help="Set process name in banner")
    parser.add_argument("-p", dest="publisher_port", default='43124',
                        help="Publisher IP port")
    parser.add_argument("-s", dest="subscriber_port", default='43125',
                        help="Subscriber IP port")

    args = parser.parse_args()

    subscription_list = args.subscription_list

    if len(subscription_list) > 1:
        subscription_list = args.subscription_list.split(',')

    kw_options = {
        'publisher_port': args.publisher_port,
        'subscriber_port': args.subscriber_port,
        'process_name': args.process_name,
        'server_ip_port': args.server_ip_port,
    }

    log = args.log.lower()
    if log == 'false':
        log = False
    else:
        log = True

    if args.back_plane_ip_address != 'None':
        kw_options['back_plane_ip_address'] = args.back_plane_ip_address

    # get the event loop
    # this is for python 3.8
    if sys.platform == 'win32':
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

    loop = asyncio.get_event_loop()

    try:
        WsGateway(subscription_list, **kw_options, event_loop=loop)
    except KeyboardInterrupt:
        for task in asyncio.Task.all_tasks():
            task.cancel()
        loop.stop()
        loop.close()
        sys.exit(0)



def signal_handler(sig, frame):
    print('Exiting Through Signal Handler')
    raise KeyboardInterrupt



# listen for SIGINT
signal.signal(signal.SIGINT, signal_handler)
signal.signal(signal.SIGTERM, signal_handler)

if __name__ == '__main__':
    ws_gateway()



from setuptools import setup

with open('pypi_desc.md') as f:
    long_description = f.read()

setup(
    name='s3-extend',
    version='1.19',
    packages=[
      's3_extend',
      's3_extend.gateways'
    ],
    install_requires=[
        'python-banyan>=3.9',
        'pymata-express>=1.11',
        'pymata_rh',
        'pymata-cpx',
        'tmx-pico-aio',
        'telemetrix-aio'
    ],

    entry_points={
        'console_scripts': [
            's3a = s3_extend.s3a:s3ax',
            's3c = s3_extend.s3c:s3cx',
            's3e = s3_extend.s3e:s3ex',
            's3p = s3_extend.s3p:s3px',
            's3r = s3_extend.s3r:s3rx',
            's3rh = s3_extend.s3rh:s3rhx',
            's3rp = s3_extend.s3rp:s3rpx',
            'ardgw = s3_extend.gateways.arduino_gateway:arduino_gateway',
            'cpxgw = s3_extend.gateways.cpx_gateway:cpx_gateway',
            'espgw = s3_extend.gateways.esp8266_gateway:esp8266_gateway',
            'pbgw = s3_extend.gateways.picoboard_gateway:picoboard_gateway',
            'rpigw = s3_extend.gateways.rpi_gateway:rpi_gateway',
            'rhgw = s3_extend.gateways.robohat_gateway:robohat_gateway',
            'rpgw = s3_extend.gateways.rpi_pico_gateway:rpi_pico_gateway',
            'wsgw = s3_extend.gateways.ws_gateway:ws_gateway',
        ]
    },

    url='https://github.com/MrYsLab/s3-extend',
    license='GNU Affero General Public License v3 or later (AGPLv3+)',
    author='Alan Yorinks',
    author_email='MisterYsLab@gmail.com',
    description='A Non-Blocking Event Driven Applications Framework',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=['Scratch3', 'Arduino', 'ESP-8266', 'Raspberry Pi'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Other Environment',
        'Intended Audience :: Education',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.7',
        'Topic :: Education',
        'Topic :: Software Development',
        'Topic :: System :: Hardware'
    ],
)
